(***********************************************************************************
**  HanoiIterativ.mod: Iterative Lsung zum (hoffentlich bekannten) Hanoi-Problem  **
**                                                                                 **
** Programmiert von Michael Khn, MKSoft, Software fr den Unterricht, Feb. 2001   **
** www.bm-kuehn.de, kuehn@c-g-j.de                                                 **
**                                                                                 **
** Dieses Programm darf im Unterricht verwendet werden, wenn keine nderungen      **
** vorgenommen wurden.                                                             **
************************************************************************************)
(************************************************************************************
** Problemlsung iterativ:                                                         **
**                                                                                 **
** 1. Jeder zweite Zug besteht aus dem Bewegen der kleinsten Scheibe. Diese wan-   **
**    dert von Stang 0 zu Stange 1, dann zu Stange 2, dann wieder zu Stange 0 usw. **
**    (Bei ungeradem n andersrum: Stange 0, Stange 2, Stange 1 usw.).              **
** 2. Bei den anderen Zgen muss man sich die Stangen ansehen, auf denen die       **
**    kleinste Scheibe NICHT liegt. Liegen auf beiden Stangen Scheiben, so wandert **
**    die kleinere auf die grere, sonst von der Stange mit Scheibe auf die       **
**    freie.                                                                       **
** 3. Dies wird wiederholt, bis man fertig ist, d.h. bis alle Scheiben auf Stange  **
**    Nr. 2 sind.                                                                  **
**                                                                                 **
** Die Stangen wurden 0,1,2 genannt, um leichter den MOD-Operator verwenden zu     **
** knnen (siehe Schritt 1).                                                       **
************************************************************************************)
MODULE HanoiIterativ; 
IMPORT D:=Display;

CONST max=10;
      mitKontrollausgabe = TRUE;
      mitWartenNachJedemSchritt = FALSE;
  
TYPE stabTyp = ARRAY 3, max+1 OF INTEGER;
    (* stab[0,2]=3 heit, dass auf Stab 0 an 2. unterster Stelle die dritt
       kleinste Scheibe liegt.
       stab[i,j]=k -> i: Stabnummer (0,1,2)
                      j: Position auf Stab (1=unten, n bzw. max=ganz oben)
                      k: Nummer der Scheibe an dieser Position (1=ganz klein)
     *)
     anzTyp  = ARRAY 3 OF INTEGER;
     (* anz[i] ist die Anzahl der Scheiben auf Stab i *)

(* globale Variable *)
VAR zuege: INTEGER;  (* Anzahl der Zge, wird mitgezhlt *)

PROCEDURE warteAufTaste;
 (* lscht den Tastaturpuffer und wartet *)
VAR ch: CHAR;
BEGIN
  WHILE D.KeyPressed() DO
    ch := D.ReadKey();
  END;
  REPEAT UNTIL D.KeyPressed();
  ch := D.ReadKey(); 
END warteAufTaste;

PROCEDURE kontrollAusgabe(n: INTEGER; stab: stabTyp; anz:anzTyp);
VAR i,j: INTEGER;
BEGIN
  IF mitKontrollausgabe
  THEN
    D.WriteLn;
    D.WriteStr("Kontrollausgabe:");
    D.WriteLn;
    FOR i:=0 TO 2 DO
      D.WriteStr("Stab Nr. ");
      D.WriteInt(i,1);
      D.WriteStr(": ");
      FOR j:=1 TO anz[i] DO
        D.WriteInt(stab[i,j],3);
      END;
      D.WriteLn;
      D.WriteLn;
    END;
    warteAufTaste;
  END;
END kontrollAusgabe;

PROCEDURE initialisiere(n: INTEGER; VAR stab: stabTyp; 
                        VAR anz: anzTyp; VAR klein: INTEGER);
VAR i,j: INTEGER;
BEGIN
  FOR i:=0 TO 2 DO (* erstmal alles leer *)
    FOR j:=1 TO max DO
      stab[i,j]:=0;
    END;
  END;
  (* Jetzt Stab 0 mit Scheiben belegen *)
  FOR j:=1 TO n DO
    stab[0,j]:=n-j+1;
  END;
  anz[0]:=n;
  anz[1]:=0;
  anz[2]:=0;
  klein:=0;
  kontrollAusgabe(n,stab,anz);
END initialisiere;

PROCEDURE versetzeKleinsteScheibe(n: INTEGER; VAR stab: stabTyp; 
                                  VAR anz: anzTyp; VAR klein: INTEGER);
BEGIN
  D.WriteStr("Versetze die kleinste Scheibe von Stab ");
  D.WriteInt(klein,1);
  
  stab[klein,anz[klein]]:=0;
  DEC(anz[klein]);
  IF ODD(n) 
  THEN klein := (klein-1) MOD 3;
  ELSE klein := (klein+1) MOD 3;
  END;
  INC(anz[klein]);
  stab[klein,anz[klein]]:=1;
  
  D.WriteStr(" nach Stab ");
  D.WriteInt(klein,1);
  D.WriteLn;
  INC(zuege);
  IF mitWartenNachJedemSchritt OR ((zuege MOD 20=0) & ~mitKontrollausgabe) 
  THEN warteAufTaste 
  END;
END versetzeKleinsteScheibe;

PROCEDURE versetzeAndereScheibe(n: INTEGER; VAR stab: stabTyp; 
                                  VAR anz: anzTyp; VAR klein: INTEGER);
VAR stabA,stabB,hilf,scheibeNr: INTEGER; 
(* stabA, stabB: diejenigen Stbe, auf denen NICHT die kleinste Scheibe ist *)
BEGIN
  stabA:=(klein+1) MOD 3;
  stabB:=(stabA+1) MOD 3;
  (* Auf stabA soll die kleinere der beiden Scheiben liegen, sofern 
     berhaupt auf beiden eine liegt. U.U. Tausch                    *)
  IF (anz[stabA]=0) OR ((stab[stabA,anz[stabA]] > stab[stabB,anz[stabB]]) & (anz[stabB] # 0))
  THEN
    hilf:=stabA;
    stabA:=stabB;
    stabB:=hilf;
  END;
  (* jetzt ist auf stabA die kleinere der beiden Scheiben, d.h. man kann
  von stabA eine Scheibe auf stabB versetzen.                             *)
  scheibeNr:=stab[stabA,anz[stabA]];
  stab[stabA,anz[stabA]]:=0;
  DEC(anz[stabA]);
  INC(anz[stabB]);
  stab[stabB,anz[stabB]]:=scheibeNr;
  D.WriteStr("Versetze die Scheibe Nr. ");
  D.WriteInt(scheibeNr,1);
  D.WriteStr(" von Stab ");
  D.WriteInt(stabA,1);
  D.WriteStr(" nach Stab ");
  D.WriteInt(stabB,1);
  D.WriteLn;
  INC(zuege);
  IF mitWartenNachJedemSchritt OR ((zuege MOD 20=0) & ~mitKontrollausgabe) 
  THEN warteAufTaste 
  END;
END versetzeAndereScheibe;

PROCEDURE fertig(anz: anzTyp): BOOLEAN;
BEGIN
  RETURN (anz[0]=0) & ((anz[1]=0) OR (anz[2]=0));
END fertig;
                                    

PROCEDURE hanoi(n: INTEGER); 
VAR stab: stabTyp;
    anz: anzTyp;
    klein: INTEGER; (* 0,1 oder 2: Stabnummer, wo gerade die kleinste 
                     Scheibe ist *)
BEGIN
  initialisiere(n,stab,anz,klein);
  versetzeKleinsteScheibe(n,stab,anz,klein);
  kontrollAusgabe(n,stab,anz);
  REPEAT
    versetzeAndereScheibe(n,stab,anz,klein);
    versetzeKleinsteScheibe(n,stab,anz,klein);
    kontrollAusgabe(n,stab,anz);
  UNTIL fertig(anz);
END hanoi;

PROCEDURE ProgMain*;
VAR n: INTEGER;
    code: CHAR;
BEGIN
  (* Eingabe von n *)
  D.WriteStr("Turm von Hanoi mit n Scheiben (0 < n < ");
  D.WriteInt(max-1,1);
  D.WriteStr(" )");
  D.WriteLn;
  REPEAT
    D.WriteStr("n = ");
    D.ReadInt(n,2,code);
    D.WriteLn;
  UNTIL (code=D.ENTER) & (n>0) & (n<max);
  D.ClrScr;

  (* berschrift *)
  D.WriteStr("Turm von Hanoi mit ");
  D.WriteInt(n,1);
  D.WriteStr(" Scheiben.");
  D.WriteLn;
  D.WriteLn;
  
  (* Lsung und Ausgabe *)

  hanoi(n); 
  
  (* Auswertung der Anzahl der Zge *)
  D.WriteLn;
  D.WriteStr("Insgesamt ");
  D.WriteInt(zuege,1);
  D.WriteStr(" Zge waren ntig.");
  D.WriteLn;
  warteAufTaste;
END ProgMain;

END HanoiIterativ.
